using System;
using System.Globalization;
using System.Threading.Tasks;
using FarmmapsApi.Models;
using FarmmapsApi.Services;
using Microsoft.Extensions.Logging;
using static FarmmapsApi.Extensions;
using static FarmmapsApiSamples.Constants;

namespace FarmmapsNbs
{
    public class NitrogenService
    {
        private readonly ILogger<NitrogenService> _logger;
        private readonly FarmmapsApiService _farmmapsApiService;
        private readonly GeneralService _generalService;

        public NitrogenService(ILogger<NitrogenService> logger, FarmmapsApiService farmmapsApiService,
            GeneralService generalService)
        {
            _logger = logger;
            _farmmapsApiService = farmmapsApiService;
            _generalService = generalService;
        }

        public async Task<Item> CreateTargetNItem(Item cropfieldItem)
        {
            var itemRequest = new ItemRequest()
            {
                ParentCode = cropfieldItem.ParentCode,
                ItemType = USERINPUT_ITEMTYPE,
                Name = "TargetN"
            };
            return await _farmmapsApiService.CreateItemAsync(itemRequest);
        }
        
        /// <summary>
        /// Calculates TargetN, makes the assumption the cropfield and user.input(targetn) item have the same parent
        /// </summary>
        /// <param name="cropfieldItem">The cropfield to base the calculations on</param>
        /// <param name="targetYield">The target yield input for the TargetN calculation</param>
        /// <param name="plantingDate">The date the crop is planted</param>
        /// <param name="measurementDate">The date the measurements are taken</param>
        /// <returns>The TargetN</returns>
        public async Task<double> CalculateTargetN(Item cropfieldItem, Item targetNItem, DateTime plantingDate, 
            DateTime measurementDate, string inputType, string purposeType, int targetYield)
        {
            var nbsTargetNRequest = new TaskRequest {TaskType = VRANBS_TASK};
            nbsTargetNRequest.attributes["operation"] = "targetn";
            nbsTargetNRequest.attributes["inputCode"] = targetNItem.Code;
            nbsTargetNRequest.attributes["plantingDate"] = plantingDate.ToString();
            nbsTargetNRequest.attributes["measurementDate"] = measurementDate.ToString();
            nbsTargetNRequest.attributes["inputType"] = inputType;
            nbsTargetNRequest.attributes["purposeType"] = purposeType.ToLower();
            nbsTargetNRequest.attributes["targetYield"] = targetYield.ToString();
            string itemTaskCode = await _farmmapsApiService.QueueTaskAsync(cropfieldItem.Code, nbsTargetNRequest);
            
            await PollTask(TimeSpan.FromSeconds(3), async (tokenSource) =>
            {
                var itemTaskStatus = await _farmmapsApiService.GetTaskStatusAsync(cropfieldItem.Code, itemTaskCode);
                if (itemTaskStatus.IsFinished)
                    tokenSource.Cancel();
            });

            var itemTask = await _farmmapsApiService.GetTaskStatusAsync(cropfieldItem.Code, itemTaskCode);
            if(itemTask.State == ItemTaskState.Error)
            {
                _logger.LogError($"Something went wrong with task execution: {itemTask.Message}");
                return 0;
            }
            
            var item = await _farmmapsApiService.GetItemAsync(targetNItem.Code);
            if (item.Data.ContainsKey("TargetN"))
                return item.Data.Value<double>("TargetN");

            return 0;
        }
        
        /// <summary>
        /// Calculates the uptake map based on the given inputs
        /// </summary>
        /// <param name="cropfieldItem">The cropfield to base the calculations on</param>
        /// <param name="inputItem"></param>
        /// <param name="plantingDate">The date the crop is planted</param>
        /// <param name="measurementDate">The date the measurements are taken</param>
        /// <returns></returns>
        public async Task<Item> CalculateUptakeMap(Item cropfieldItem, Item inputItem, DateTime plantingDate, 
            DateTime measurementDate, string inputType = "irmi")
        {
            var nbsUptakeMapRequest = new TaskRequest {TaskType = VRANBS_TASK};
            nbsUptakeMapRequest.attributes["operation"] = "uptake";
            nbsUptakeMapRequest.attributes["inputCode"] = inputItem.Code;
            nbsUptakeMapRequest.attributes["plantingDate"] = plantingDate.ToString();
            nbsUptakeMapRequest.attributes["measurementDate"] = measurementDate.ToString();
            nbsUptakeMapRequest.attributes["inputType"] = inputType.ToLower();

            string itemTaskCode = await _farmmapsApiService.QueueTaskAsync(cropfieldItem.Code, nbsUptakeMapRequest);
            
            await PollTask(TimeSpan.FromSeconds(5), async (tokenSource) =>
            {
                var itemTaskStatus = await _farmmapsApiService.GetTaskStatusAsync(cropfieldItem.Code, itemTaskCode);
                if (itemTaskStatus.IsFinished)
                    tokenSource.Cancel();
            });
            
            var itemTask = await _farmmapsApiService.GetTaskStatusAsync(cropfieldItem.Code, itemTaskCode);
            if(itemTask.State == ItemTaskState.Error)
            {
                _logger.LogError($"Something went wrong with task execution: {itemTask.Message}");
                return null;
            }

            var itemName = "VRANbs uptake";
            var uptakeMapItem = await _generalService.FindChildItemAsync(cropfieldItem.Code,
                GEOTIFF_PROCESSED_ITEMTYPE, itemName, 
                i => i.Updated >= itemTask.Finished.GetValueOrDefault(DateTime.UtcNow) &&
                     i.Name.ToLower().Contains(itemName.ToLower()));
            if (uptakeMapItem == null)
            {
                _logger.LogError("Could not find the uptake geotiff child item under cropfield");
                return null;
            }

            return uptakeMapItem;
        }


        /// <summary>
        /// Creates the nitrogen appliance map based on given input data
        /// </summary>
        /// <param name="cropfieldItem">The cropfield to base the calculations on</param>
        /// <param name="inputItem">The farmmaps item containing the geotiff data</param>
        /// <param name="plantingDate">The date the crop is planted</param>
        /// <param name="measurementDate">The date the measurements are taken</param>
        /// <param name="inputType">The inputtype to use</param>
        /// <param name="targetN">The target nitrogen to use for the calculations</param>
        /// <returns></returns>
        public async Task<Item> CalculateApplicationMap(Item cropfieldItem, Item inputItem, DateTime plantingDate, 
            DateTime measurementDate, string inputType = "irmi", double targetN = 60.0)
        {
            var nbsApplianceMapRequest = new TaskRequest {TaskType = VRANBS_TASK};
            nbsApplianceMapRequest.attributes["operation"] = "application";
            nbsApplianceMapRequest.attributes["inputCode"] = inputItem.Code;
            nbsApplianceMapRequest.attributes["plantingDate"] = plantingDate.ToString();
            nbsApplianceMapRequest.attributes["measurementDate"] = measurementDate.ToString();
            nbsApplianceMapRequest.attributes["inputCode"] = inputItem.Code;
            nbsApplianceMapRequest.attributes["inputType"] = inputType.ToLower();
            nbsApplianceMapRequest.attributes["targetN"] = targetN.ToString(CultureInfo.InvariantCulture);
            
            string itemTaskCode = await _farmmapsApiService.QueueTaskAsync(cropfieldItem.Code, nbsApplianceMapRequest);
            
            await PollTask(TimeSpan.FromSeconds(5), async (tokenSource) =>
            {
                var itemTaskStatus = await _farmmapsApiService.GetTaskStatusAsync(cropfieldItem.Code, itemTaskCode);
                if (itemTaskStatus.IsFinished)
                    tokenSource.Cancel();
            });
            
            var itemTask = await _farmmapsApiService.GetTaskStatusAsync(cropfieldItem.Code, itemTaskCode);
            if(itemTask.State == ItemTaskState.Error)
            {
                _logger.LogError($"Something went wrong with task execution: {itemTask.Message}");
                return null;
            }

            var itemName = $"VRANbs application";
            var applianceMapItem = await _generalService.FindChildItemAsync(cropfieldItem.Code,
                GEOTIFF_PROCESSED_ITEMTYPE, itemName, 
                i => i.Updated >= itemTask.Finished.GetValueOrDefault(DateTime.UtcNow) &&
                     i.Name.ToLower().Contains(itemName.ToLower()));
            if (applianceMapItem == null)
            {
                _logger.LogError("Could not find the application map geotiff child item under cropfield");
                return null;
            }

            return applianceMapItem;
        }
    }
}